"""
Specification
Fourier Transform algorith

I. First part
1. You are given a list of discrete values from a timeseries row. It doesn't matter the what TF is used as long as the data are correct for that TF.

Typically there should be 960 periods of data to use as base for the cycle calculations.
[I use 1H MT4 data so there are 5 * 24 = 120 periods for a full trading week, respectively 960 periods are for 8 weeks.]

Another 120 periods are needed to test the validity of the model. Usually those 120 periods are the latest price data available.

2. We start by establishinf the values of T (cycle period) and N (max number of data entries divisible by T)

The max value of T should be up to 1/3 of those 960 = 320

3. Calculate a(1) & b(1) for every element of N

4. Calculate  a(1)-res & b(1)-res for T
4.1 Calcualte the coefficient for that T

5. Choose the appropriate T-s to be used: return a dictionary with format:
"T":
"a_1":
"b_1":
"coef":
5.1 Create a list of T-s for coef above certain treashold

II. Second part
11. Given a T_list a possible values for T: we need to pick up the best conbination of Ts.

12. We shall be using T_list
- all values
- all values - 1 [calculating for all combinations when a single element is removed]
- all values - 2 [same as above for every 2 elements removed]
- all values - 3 [same as above for every 3 elements removed]

13. For all variants from 12. we need to calculate the correlation versus the input data.
13.1 Correlation versus the additional data.

14. Export the list of lists results for 13.
[[[combination of periods_1], corr_960, corr_120],
[[combination of periods_2], corr_960, corr_120], ...
]

III. Third part
21. For 1) a given combination of periods and 2) start time produce: MT4 XML chart code for estimated trend curve for the next 120 periods.
- create a list of 120 - 1  unix timestamps beginning from the start time
- create every single <object> element
- combine all <object> elements
- write the result XML file
"""

import csv
import numpy as np
import pandas as pd
from math import pi, cos, sin, sqrt
from sysDef import ave_list, pd_corr, dtUniForm, uniDt, rev_date, non_intersect, ll_to_csv
from fourex import fourierExtrapolation as fE

def files():
	""" returns locations and names of the files
	used for importing and exporting data """
	return {"f_tsrow": "D:\\tmp\\trading\\patterns\\development\\harmonic-analysis\\tsrow.csv",
			"f_coef": "D:\\tmp\\trading\\patterns\\development\\harmonic-analysis\\res_coef.csv",
			'f_corr': 'D:\\tmp\\trading\\patterns\\development\\harmonic-analysis\\res_corr.csv',
			'f_xml': 'D:\\tmp\\trading\\patterns\\development\\harmonic-analysis\\res.xml'
			}


def d_min_coef():
	""" returns instruments min coefficients for
	res_ll_create() function """
	return {
			"AUDJPY": 0.15,
			"AUDNZD": 0.003,
			"NZDJPY": 0.3,
			"EURSEK": 0.02,
			"USDCAD": 0.0015,
			"USDNOK": 0.02,
			"NZDCAD": 0.0015,
			"EURUSD": 0.00175,
			"GBPCHF": 0.003,
			"GBPUSD": 0.00175,
			"CHFJPY": 0.3,
			"NZDUSD": 0.0015
			}


def xml_blocks():
	""" returns the blocks to be used
	in creating the XML code for the result file"""
	return {
			"block1":	"<object>\ntype=2\nobject_name=Trendline ",
			"block2":	"\nperiod_flags=0\ncreate_time=1593959357\ncolor=14822282\nstyle=0\nweight=3\nbackground=1\nfilling=0\nselectable=0\nhidden=0\nzorder=0\ntime_0=",
			"block2.1":	"\nperiod_flags=0\ncreate_time=1593959357\ndescription=",
			"block2.2":	"\ncolor=",
			"block2.3":	"\nstyle=0\nweight=3\nbackground=1\nfilling=0\nselectable=0\nhidden=0\nzorder=0\ntime_0=",
			"block3":	"\nvalue_0=",
			"block4":	"\ntime_1=",
			"block5":	"\nvalue_1=",
			"block6":	"\nray=0\n</object>\n"
			}


def fo_help():
	""" prints the sequence of actions
	to be performed when using fourier module """
	str_ = "1. Set up a CSV file with a single column of\nnumbers representing the time series row.\nFor basic setup you need 960 + 120 + 120 positions.\nSave that file at the location of files()['f_tsrow']\n2. Use import_double function to import the CSV file.\n3. Use res_ll_create() function to get the coefficents\nfor cycles above certain min_coef.\n4. Use gen_f_coef() function to create CSV file of the coefficients.\n5. Open the file from 4. and choose the best coefficents.\nWrite them as list=[55,62,..]\n6. Use corr_res() function to get correlation results in CSV file.\nThis calculation may take several minutes :)\nYou might need to do this several times.\ntill you get best cycles combination.\n7. When you get the best cycle combo, use fo_2().\n8. Copy/paste the XML file from 7. into MT4 profile chart."
	# print the text
	print(str_)


def N_calc(N_all, T):
	""" calcualtes N for given N_all, T """
	N_res = T * int(N_all / T)
	return N_res


def ab_calc(x, N, T, t):
	""" calculates a(1) & b(1) given:
		x - value for t-th element of timeseries vector
		N - as calcualted from N_calc
		T - period
		t - index of t-th element of timeseries vector """

	a = (2/N) * x * cos(2*pi * t / T)
	b = (2/N) * x * sin(2*pi * t / T)
	return [a, b]


def ab_res_calc(l_tsrow, N, T):
	""" calculates a(1)-res & b(1)-res for given:
		l_tsrow - list of [960] time series values
		N - as calcualted from N_calc
		T - period """

	# result variables
	a_res = 0.0
	b_res = 0.0
	# calcualtion cycle
	i = 0
	while i < N:
		ab_list = ab_calc(l_tsrow[i], N, T, i)
		a_res = a_res + ab_list[0]
		b_res = b_res + ab_list[1]
		i = i + 1
	# calculate cycle coefficient
	coef = sqrt(a_res ** 2 + b_res ** 2)
	# produce the result
	return [a_res, b_res, coef]


def res_ll_create(l_tsrow, min_coef):
	""" for a given timeseries row of values
	creates result list of lists
	above certain min_coef """

	# create result list of lists
	ll_res = []
	# calculation cycle
	T = 3
	while T < len(l_tsrow) / 3.0:
		# calcualte N
		N = N_calc( len(l_tsrow), T)
		# calculate ab_res
		l_ab_res = ab_res_calc(l_tsrow, N, T)
		# add to result file if coef is bigger than min_coef
		if l_ab_res[2] > min_coef:
			ll_res.append([T] + l_ab_res)
		T = T + 1
	# return result
	return ll_res

def proj_calc(l_ab_res, t):
	""" for a given list of ab res values
	calculates the marginal change based on
	that cycle for the projected value
	l_ab_res: [T, a_res, b_res, coef] """
	T = l_ab_res[0]
	a = l_ab_res[1]
	b = l_ab_res[2]
	# calc result using the given formula
	return a * cos(2 * pi * t / T ) + b * sin(2 * pi * t / T)


def proj_create(l_tsrow, l_proj, n_test=120):
	""" for a given list of T-s on ts row
	creates projected value row """

	# create ab result list of lists
	ll_ab_res = []
	# ab-res cycle calcualtion
	for T in l_proj:
		# calcualte N
		N = N_calc( len(l_tsrow), T)
		# calculate ab_res
		l_ab_res = ab_res_calc(l_tsrow, N, T)
		# add to ab result list of lists
		ll_ab_res.append([T] + l_ab_res)
	
	# find ts row average value
	ts_ave = ave_list(l_tsrow)
	# create list of projected values
	l_proj_res = []
	# projected values calculation cycle
	k = 0
	while k < len(l_tsrow):
		proj_val = ts_ave
		i = 0
		while i < len(ll_ab_res):		
			proj_val = proj_val + proj_calc(ll_ab_res[i], k)
			i = i + 1
		# add the projected value to the result list
		l_proj_res.append(proj_val)
		k = k + 1
	# create list of projected test values
	l_test_res = []
	# projected test values calculation cycle
	while k < (len(l_tsrow) + n_test * 2):
		proj_val = ts_ave
		i = 0
		while i < len(ll_ab_res):		
			proj_val = proj_val + proj_calc(ll_ab_res[i], k)
			i = i + 1
		# add the projected test value to the result list
		l_test_res.append(proj_val)
		k = k + 1
	# return the result list
	return [l_proj_res, l_test_res[:n_test], l_test_res[n_test:] ]


def fut_create(ll_tsrow, l_proj, n_test=120):
	""" for a given list of T-s on ts row
	series with calc and test data
	creates future value row, assuming both
	test data & future data of 120 periods
	EDIT: Made for 2 test datasets """

	# create ab result list of lists
	ll_ab_res = []
	# combine calc and test tsrow data
	l_tsrow = ll_tsrow[0] + ll_tsrow[1] + ll_tsrow[2]
	# ab-res cycle calcualtion
	for T in l_proj:
		# calcualte N
		N = N_calc( len(l_tsrow), T)
		# calculate ab_res
		l_ab_res = ab_res_calc(l_tsrow, N, T)
		# add to ab result list of lists
		ll_ab_res.append([T] + l_ab_res)
	#print(ll_ab_res)

	# find ts row average value
	ts_ave = ave_list(l_tsrow)
	# create list of future values
	l_fut_res = []
	# projected future values calculation cycle
	k = len(l_tsrow)
	while k < len(l_tsrow) + n_test:
		proj_val = ts_ave
		i = 0
		while i < len(ll_ab_res):		
			proj_val = proj_val + proj_calc(ll_ab_res[i], k)
			i = i + 1
		# add the projected test value to the result list
		l_fut_res.append(proj_val)
		k = k + 1
	# return the result list
	return l_fut_res


def fut_min_max(l_fut_res, ll_xml):
	""" for a given future ts row and
	XML list of lists gives the datetimes of
	the minimum and the maximum	"""
	
	# create pandas dataseries object
	ds_fut = pd.Series(l_fut_res)
	# get the min and the max of l_fut_res
	min_ = ds_fut.min()
	max_ = ds_fut.max()
	# list with the indexes of the min and the max
	l_ind = []
	for k in [min_, max_]:
		l_ind.append(l_fut_res.index(k))
	# get the datetime for the min and for the max
	l_datetime = []
	for k in l_ind:
		l_datetime.append(uniDt(ll_xml[k][1]))
	# print the result
	l_str = ["minimum",'maximim']
	i = 0
	while i < 2:
		print("The " + l_str[i] + ' is at ' + l_datetime[i])
		i = i + 1


def corr_res(ll_tsrow, l_proj, f_corr):
	""" returns a number of combinations
	that give correlation results for given
	periods [l_proj], projected and test values
	and writes those to CSV file """
	
	# set up the result file
	ll_corr_res = []
	# create variants of l_proj
	# 1) calc orig l_proj
	corr_0 = pd_corr(ll_tsrow[0], proj_create(ll_tsrow[0], l_proj)[0])
	corr_1 = pd_corr(ll_tsrow[1], proj_create(ll_tsrow[0], l_proj)[1])
	if len(ll_tsrow) == 3:
		corr_2 = pd_corr(ll_tsrow[2], proj_create(ll_tsrow[0], l_proj)[2])
	# add those to ll_corr_res
	if len(ll_tsrow) == 3:
		ll_corr_res.append([l_proj, ["full"], corr_0, corr_1, corr_2])
	else:
		ll_corr_res.append([l_proj, ["full"], corr_0, corr_1])
	# 2) calc all l_proj combinations when removing one element at a time
	for k in list(range(len(l_proj))):
		# create new list instance for every combination
		list1 = []
		for j in l_proj:
			list1.append(j)
		# remove one element from the combination list
		del_ele = list1[k]
		del list1[k]
		# calculate corr coefficent for the combination list
		corr_0 = pd_corr(ll_tsrow[0], proj_create(ll_tsrow[0], list1)[0])
		corr_1 = pd_corr(ll_tsrow[1], proj_create(ll_tsrow[0], list1)[1])
		if len(ll_tsrow) == 3:
			corr_2 = pd_corr(ll_tsrow[2], proj_create(ll_tsrow[0], list1)[2])
		# add those to ll_corr_res
		if len(ll_tsrow) == 3:
			ll_corr_res.append([list1, [del_ele], corr_0, corr_1, corr_2])
		else:
			ll_corr_res.append([list1, [del_ele], corr_0, corr_1])
	# 3) calc all l_proj combinations when removing two elements at a time
	for k in list(range(len(l_proj))):
		n = k + 1
		while n < len(l_proj):
			# create new list instance for every combination
			list2 = []
			for j in l_proj:
				list2.append(j)
			# remove two elements from the combination list
			del_1 = list2[n]
			del list2[n]
			# first we needed to remove the cycle with the higher index
			del_2 = list2[k]
			del list2[k]
			# calculate corr coefficent for the combination list
			corr_0 = pd_corr(ll_tsrow[0], proj_create(ll_tsrow[0], list2)[0])
			corr_1 = pd_corr(ll_tsrow[1], proj_create(ll_tsrow[0], list2)[1])
			if len(ll_tsrow) == 3:
				corr_2 = pd_corr(ll_tsrow[2], proj_create(ll_tsrow[0], list2)[2])
			# add those to ll_corr_res
			if len(ll_tsrow) == 3:
				ll_corr_res.append([list2, [del_1, del_2], corr_0, corr_1, corr_2])
			else:
				ll_corr_res.append([list2, [del_1, del_2], corr_0, corr_1])
			n = n + 1

	# export result file to CSV file
	ll_to_csv(ll_corr_res, f_corr)
	"""
	with open(f_corr, "w", newline="") as f:
		writer = csv.writer(f)
		writer.writerows(ll_corr_res)
	"""
	print("Result file written!")
	# return results list of lists
	return ll_corr_res
	

def corr_one_up(ll_tsrow, l_proj, f_corr):
	""" similar to corr_res but combines
	l_proj with every other cycle and 
	writes the result to CSV file """

	# set up the result file
	ll_corr_res = []
	# create variants of l_proj
	# 1) calc orig l_proj
	corr_0 = pd_corr(ll_tsrow[0], proj_create(ll_tsrow[0], l_proj)[0])
	corr_1 = pd_corr(ll_tsrow[1], proj_create(ll_tsrow[0], l_proj)[1])
	if len(ll_tsrow) == 3:
		corr_2 = pd_corr(ll_tsrow[2], proj_create(ll_tsrow[0], l_proj)[2])
	# add those to ll_corr_res
	if len(ll_tsrow) == 3:
		ll_corr_res.append([l_proj, ['full'], corr_0, corr_1, corr_2])
	else:
		ll_corr_res.append([l_proj, ['full'], corr_0, corr_1])
	# 2) combinations with all non included elements
	l_one_up = list(range(3, 320))
	l_one_up = non_intersect(l_one_up, l_proj)
	i = 0
	while i < len(l_one_up):
		list1 = l_proj + [l_one_up[i]]
		# calculate corr coefficent for the combination list
		corr_0 = pd_corr(ll_tsrow[0], proj_create(ll_tsrow[0], list1)[0])
		corr_1 = pd_corr(ll_tsrow[1], proj_create(ll_tsrow[0], list1)[1])
		if len(ll_tsrow) == 3:
			corr_2 = pd_corr(ll_tsrow[2], proj_create(ll_tsrow[0], list1)[2])
		# add those to ll_corr_res
		if len(ll_tsrow) == 3:
			ll_corr_res.append([list1, [l_one_up[i]], corr_0, corr_1, corr_2])
		else:
			ll_corr_res.append([list1, [l_one_up[i]], corr_0, corr_1])
		i = i + 1

	# export result file to CSV file
	ll_to_csv(ll_corr_res, f_corr)
	print("Result file written!")
	# return results list of lists
	return ll_corr_res


def multi_fE(ll_tsrow, n_fut=120, n_param=26, line_fit=1, pablo_opt=0):
	""" uses fE from fourex to test
	the validity of the test data
	for different harmonics [n_harm] """

	np_tsrow_0 = np.array(ll_tsrow[0])
	len_0 = len(ll_tsrow[0])
	len_1 = len_0 + n_fut
	ll_res = []
	for n_harm in list(range(3,25)):
		l_proj = fE(np_tsrow_0, n_fut * 2, n_harm, n_param, line_fit, pablo_opt).tolist()
		l_fut_1 = l_proj[len_0:len_1]
		l_fut_2 = l_proj[len_1:]
		corr_0 = pd_corr(ll_tsrow[0], l_proj[:len_0])
		corr_1 = pd_corr(ll_tsrow[1], l_fut_1)
		corr_2 = pd_corr(ll_tsrow[2], l_fut_2)
		#print([n_harm, corr_1, corr_2])
		ll_res.append([n_harm, corr_0, corr_1, corr_2])
	# export result file to CSV file
	f_corr = files()['f_corr']
	ll_to_csv(ll_res, f_corr)
	print("Correlation results file written!")
	# return the result
	return ll_res


def best_fE(ll_tsrow, n_harm, n_fut=120, n_param=26, line_fit=1, pablo_opt=0):
	""" for the best harmonic calculated
	in multi_fE, gives the future ts row
	n_fut - number of periods for test/future calculation
	n_harm - best harmonic """

	len_0 = len(ll_tsrow[0])
	np_tsrow_0 = np.array(ll_tsrow[0])
	np_proj = fE(np_tsrow_0, n_fut * 3, n_harm, n_param, line_fit, pablo_opt)
	return np_proj.tolist()[(len_0 + n_fut * 2):]


def best_corr(ll_corr_res):
	""" for a given list of list with
	projected and test corr coefficients
	returns the best combination, 
	sorted first by projected, then by test corr """
	
	# list of headers for pandas dataframe
	l_header = ['comb_list','proj_corr','test_corr']
	# create pandas dataframe
	df_corr = pd.DataFrame(ll_corr_res, columns=l_header)
	# sort descending by corr columns
	df_corr = df_corr.sort_values(by=['proj_corr','test_corr'], ascending=False)
	# best combinatio to result list
	l_res = df_corr.values.tolist()[0]
	# return best combination
	return l_res[0]


def import_tsrow(f_csv, test_lim=960):
	""" importing a timeseries row from .csv file
	using file location from files()
	test_lim - index of the first value of the test data
	output: l_tsrow, format: [ l_calc, l_test] """

	# create the result list
	l_calc =[]
	# opening the csv file
	with open(f_csv, newline='') as f:
		reader = csv.reader(f)
		data = list(reader)
	# converting calc data into numbers
	i = 0
	while i < test_lim:
		l_calc.append(float(data[i][0]))
		i = i + 1
	# create the test list
	l_test = []
	# converting test data into numbers
	while i < len(data):
		l_test.append(float(data[i][0]))
		i = i + 1
	# create the result list
	ll_tsrow = [ l_calc, l_test]
	# produce the result
	return ll_tsrow


def import_double(f_csv, test_lim=960):
	""" wrapper for import_tsrow() for 
	using two test sets of 120 periods;
	the CSV file should have 1200 records """
	ll_tsrow = import_tsrow(f_csv, test_lim)
	l_test_1 = ll_tsrow[1][:120]
	l_test_2 = ll_tsrow[1][120:]
	# return the result
	return [ll_tsrow[0], l_test_1, l_test_2]


def gen_xml_list(l_fut_res, st_date, dtForm="%d.%m.%Y", uni_step=3600):
	""" creats results xml list when given
	l_fut_res - list of future values of ts row,
	st_date - datetime for the start of the week, '13.7.2020',
	uni_step - step for unix timestamp, default = 1 hour = 3600 """
	
	# convert date to uni_timestamp plus 2 or 3 hours correction +
	# + due to difference to UTC time
	uni_date = int(dtUniForm(st_date, dtForm)) + 3 * 3600
	# create l_uni_date
	l_uni_date = []
	for k in list(range(len(l_fut_res))):
		new_uni = uni_date + uni_step * k
		l_uni_date.append(new_uni)
	# create list with object_names starting with the starting date 
	l_obj_names = list(range(1000,1000 + len(l_fut_res) - 1))
	rev_date_ = rev_date(st_date)
	for k in l_obj_names:
		l_obj_names[l_obj_names.index(k)] = rev_date_ + str(k)
	# create result list of lists
	ll_xml = []
	# result cycle
	for k in list(range(len(l_obj_names))):
		ll_xml.append([l_obj_names[k], str(l_uni_date[k]), str(l_fut_res[k]), str(l_uni_date[k+1]), str(l_fut_res[k+1])])
	# return result
	return ll_xml


def gen_xml_file(ll_xml, f_res):
	""" creats results xml file """
	str_res = ''
	for k in ll_xml:
		str_res = str_res + xml_blocks()["block1"] + k[0] + xml_blocks()["block2"] + k[1] + xml_blocks()["block3"] + k[2] + xml_blocks()["block4"] + k[3] + xml_blocks()["block5"] + k[4] + xml_blocks()["block6"]
	text_file = open(f_res, "w")
	text_file.write(str_res)
	text_file.close()
	print("XML file written!")


def gen_f_coef(ll_res, f_res):
	""" creats coefficients results file """
	with open(f_res, "w", newline="") as f:
		writer = csv.writer(f)
		writer.writerows(ll_res)
	print("Result file written!")


def fo_2(ll_tsrow, l_proj, st_date):
	""" combines fut_create(), gen_xml_list() 
	and gen_xml_file into single function """
	l_fut_res = fut_create(ll_tsrow, l_proj)
	ll_xml = gen_xml_list(l_fut_res, st_date)
	f_res = files()['f_xml']
	gen_xml_file(ll_xml, f_res)
